from .classwise import ClassWisePredictor
from .cluster import ClusterPredictor
from .split import SplitPredictor
from .weight import WeightedPredictor

#
# Where ClassWisePredictor and ClusterPredictor differ from InductivePredictor are the ways to calculate thresholds.
#
